import pytest


def test_can_lock(poolcounter, clients, lock_type):
    """Can lock"""
    client = clients.get(1)
    client.send('%s l 1 1 1' % lock_type)
    assert client.receive() == 'LOCKED'


def test_can_lock_and_release(poolcounter, clients, lock_type):
    """Can lock and release"""
    client = clients.get(1)
    client.send('%s l 1 1 1' % lock_type)
    assert client.receive() == 'LOCKED'
    client.send('RELEASE')
    assert client.receive() == 'RELEASED'


def test_can_lock_and_release_and_relock(poolcounter, clients, lock_type):
    """Can lock and release and relock"""
    client = clients.get(1)
    client.send('%s l 1 1 1' % lock_type)
    assert client.receive() == 'LOCKED'
    client.send('RELEASE')
    assert client.receive() == 'RELEASED'
    client.send('%s l 1 1 1' % lock_type)
    assert client.receive() == 'LOCKED'


def test_can_lock_and_release_and_relock_and_unlock(
        poolcounter, clients, lock_type):
    """Can lock and release and relock and unlock"""
    client = clients.get(1)
    client.send('%s l 1 1 1' % lock_type)
    assert client.receive() == 'LOCKED'
    client.send('RELEASE')
    assert client.receive() == 'RELEASED'
    client.send('%s l 1 1 1' % lock_type)
    assert client.receive() == 'LOCKED'
    client.send('RELEASE')
    assert client.receive() == 'RELEASED'


def test_release_without_lock(poolcounter, clients):
    """RELEASEing without holding any locks warns"""
    client = clients.get(1)
    client.send('RELEASE')
    assert client.receive() == 'NOT_LOCKED'


def test_locking_holding_four_locks(poolcounter, clients, lock_type):
    """Locking while holding four locks warns"""
    client = clients.get(1)
    for i in range(0, 4):
        client.send('%s l%s 1 1 1' % (lock_type, str(i).encode()))
        assert client.receive() == 'LOCKED'
    client.send('%s l5 1 1 1' % lock_type)
    assert client.receive() == 'LOCK_HELD'


def test_locking_no_timeout(poolcounter, clients, lock_type):
    """Locking with no timeout LOCKs"""
    client = clients.get(1)
    client.send('%s l 1 1 0' % lock_type)
    assert client.receive() == 'LOCKED'


def test_locking_timeout_left_off(poolcounter, clients, lock_type):
    """Locking with timeout left off LOCKs"""
    client = clients.get(1)
    client.send('%s l 1 1' % lock_type)
    assert client.receive() == 'LOCKED'


def test_locking_garbage_timeout(poolcounter, clients, lock_type):
    """Locking with garbage timeout LOCKs"""
    client = clients.get(1)
    client.send('%s l 1 1 garbage' % lock_type)
    assert client.receive() == 'LOCKED'


def test_same_lock_twice(poolcounter, clients, lock_type):
    """The same lock can be locked twice"""
    client = clients.get(1)
    client.send('%s l 2 2 1' % lock_type)
    assert client.receive() == 'LOCKED'
    client.send('%s l 2 2 1' % lock_type)
    assert client.receive() == 'LOCKED'


def test_same_lock_twice_timeout(poolcounter, clients, lock_type):
    """Locking the same lock twice produces TIMEOUT with non-zero timeout"""
    client = clients.get(1)
    client.send('%s l 1 2 1' % lock_type)
    assert client.receive() == 'LOCKED'
    client.send('%s l 1 2 1' % lock_type)
    assert client.receive() == 'TIMEOUT'


def test_same_lock_twice_zero_timeout(poolcounter, clients, lock_type):
    """Locking the same lock twice produces TIMEOUT with non-zero timeout"""
    client = clients.get(1)
    client.send('%s l 1 2 0' % lock_type)
    assert client.receive() == 'LOCKED'
    client.send('%s l 1 2 0' % lock_type)
    assert client.receive() == 'TIMEOUT'


def test_same_lock_twice_queue_full(poolcounter, clients, lock_type):
    """Locking the same lock twice produces QUEUE_FULL"""
    client = clients.get(1)
    client.send('%s l 1 1 1' % lock_type)
    assert client.receive() == 'LOCKED'
    client.send('%s l 1 1 1' % lock_type)
    assert client.receive() == 'QUEUE_FULL'


def test_timeout_doesnt_unlock(poolcounter, clients, lock_type):
    """Timing out doesn't unlock locks already held"""
    client = clients.get(1)
    client.send('%s l 1 2 1' % lock_type)
    assert client.receive() == 'LOCKED'
    client.send('%s l 1 2 0' % lock_type)
    assert client.receive() == 'TIMEOUT'
    client.send('%s l 1 2 0' % lock_type)
    assert client.receive() == 'TIMEOUT'
    client.send('RELEASE')
    assert client.receive() == 'RELEASED'


@pytest.mark.parametrize('timeout', [b'0', b'1'])
def test_timeout_doesnt_consume(poolcounter, clients, lock_type, timeout):
    """Timing out doesn't consume a connection's locks"""
    client = clients.get(1)
    client.send('%s l 1 2 %s' % (lock_type, timeout))
    assert client.receive() == 'LOCKED'
    for i in range(0, 6):
        client.send('%s l 1 2 %s' % (lock_type, timeout))
        assert client.receive() == 'TIMEOUT'


def test_no_queue_slots_not_consumed(poolcounter, clients, lock_type):
    """Running out of queue slots doesn't consume a connection's locks"""
    client = clients.get(1)
    client.send('%s l 1 1 1' % lock_type)
    assert client.receive() == 'LOCKED'
    for i in range(0, 6):
        client.send('%s l 1 1 1' % lock_type)
        assert client.receive() == 'QUEUE_FULL'


def test_release_not_consumed(poolcounter, clients, lock_type):
    """RELEASEing doesn't consume a connection's locks"""
    client = clients.get(1)
    for i in range(0, 7):
        client.send('%s l 1 1 1' % lock_type)
        assert client.receive() == 'LOCKED'
        client.send('RELEASE')
        assert client.receive() == 'RELEASED'


def test_disconnect_release(poolcounter, clients, lock_type):
    """Disconnecting releases locks and you can reconnect and get them back"""
    client = clients.get(1)
    client.send('%s l 1 10 1' % lock_type)
    assert client.receive() == 'LOCKED'
    # Disconnect client, reconnect anew
    client.close()
    client = clients.get(1)
    client.send('%s l 1 10 10' % lock_type)
    assert client.receive() == 'LOCKED'



import os
import pytest
import socket
import subprocess
import time

socket.setdefaulttimeout(5)


class ClientPool:
    """hands out Client instances and tears them down"""
    def __init__(self):
        self.local_ports = set()
        self.clients = set()

    def get(self, count):
        ret = []
        for i in range(1, count + 1):
            try:
                local_port = self.local_ports.pop()
            except KeyError:
                local_port = 0
            client = Client(str(i), local_port)
            ret.append(client)
            self.clients.add(client)

        if count == 1:
            return ret[0]
        else:
            return ret

    def teardown(self):
        for client in self.clients:
            port = client.close()
            if port:
                # Recycle!
                self.local_ports.add(port)
        self.clients.clear()


class Client:
    """Wrapper around socket"""
    host = '127.0.0.1'
    port = 7531

    def __init__(self, name, local_port=0):
        self.name = name
        self.local_port = local_port
        self._socket = None

    @property
    def socket(self):
        if self._socket is None:
            tries = 5
            connected = False
            while not connected:
                self._socket = socket.socket()
                # TODO: is the getaddrinfo indirection necessary?
                remote = socket.getaddrinfo(self.host, self.port)
                local = socket.getaddrinfo(self.host, self.local_port)
                self._socket.setsockopt(
                    socket.SOL_SOCKET, socket.SO_REUSEADDR, 1
                )
                try:
                    # We try and reuse the local port if we can. The only
                    # way to do that is is to bind on 0, copy the ephemeral
                    # port we're given, and then rebind on that the next
                    # time we open the socket.
                    self._socket.bind(local[0][4])
                    self._socket.connect(remote[0][4])
                    connected = True
                except (ConnectionRefusedError, OSError):
                    tries -= 1
                    if tries <= 0:
                        raise
                    # Wait for poolcounter to start?
                    time.sleep(2)
                    # else try again!

        return self._socket

    def send(self, command):
        print('%s -> ' % self.name + command)
        self.socket.send(command.encode() + b'\n')

    def receive(self):
        recv = self.socket.recv(4096).strip().decode()
        print('%s <- ' % self.name + recv)
        return recv

    def close(self):
        if self._socket:
            # Save the local_port for next time
            self.local_port = self._socket.getsockname()[1]
            # TODO: If _socket isn't connected, trying to shut
            # it down will error again
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
            self._socket = None
        return self.local_port


def find_poolcounterd():
    # env POOLCOUNTERD can point to the binary if you want to run tests
    # against an installed version
    path = os.environ.get(
        'POOLCOUNTERD',
        os.path.join(os.path.dirname(os.path.dirname(__file__)), 'poolcounterd')
    )
    if not os.path.exists(path):
        pytest.skip('%s does not exist' % path)
    return path


@pytest.fixture(scope='session')
def poolcounter():
    path = find_poolcounterd()
    daemon = subprocess.Popen([path])
    yield daemon
    daemon.terminate()


@pytest.fixture()
def poolcounter_path():
    return find_poolcounterd()


# Use a singleton ClientPool to keep local_ports global
pool = ClientPool()


@pytest.fixture()
def clients():
    yield pool
    pool.teardown()


def pytest_generate_tests(metafunc):
    """parameterize lock_type if it's listed"""
    if 'lock_type' in metafunc.fixturenames:
        metafunc.parametrize('lock_type', ['ACQ4ME', 'ACQ4ANY'])



import pytest
import subprocess


def test_garbage(poolcounter, clients):
    """Bad commands result in error"""
    client = clients.get(1)
    client.send('GARBAGE')
    assert client.receive() == 'ERROR BAD_COMMAND'


def test_locking_just_lockname(poolcounter, clients, lock_type):
    """Locking with just lock name results in error"""
    client = clients.get(1)
    client.send('%s l' % lock_type)
    assert client.receive() == 'ERROR BAD_SYNTAX'


def test_locking_no_workers(poolcounter, clients, lock_type):
    """Locking with no workers results in error"""
    client = clients.get(1)
    client.send('%s l 0 1 1' % lock_type)
    assert client.receive() == 'ERROR BAD_SYNTAX'


def test_locking_no_queue(poolcounter, clients, lock_type):
    """Locking with no queue results in error"""
    client = clients.get(1)
    client.send('%s l 1 0 1' % lock_type)
    assert client.receive() == 'ERROR BAD_SYNTAX'


def test_locking_non_integers(poolcounter, clients, lock_type):
    """Locking with no queue results in error"""
    client = clients.get(1)
    client.send('%s l GARBAGE 0 1' % lock_type)
    assert client.receive() == 'ERROR BAD_SYNTAX'


@pytest.mark.xfail  # flaky, sometimes timeouts in final recv
def test_locking_while_waiting(poolcounter, clients, lock_type):
    """Locking while waiting is an error"""
    client1, client2 = clients.get(2)
    client2.send('%s l 1 10 10' % lock_type)
    assert client2.receive() == 'LOCKED'
    client1.send('%s l 1 10 10' % lock_type)
    client1.send('%s l 1 10 10' % lock_type)
    assert client1.receive() == 'ERROR WAIT_FOR_RESPONSE'


def test_invalid_listen(poolcounter_path):
    with pytest.raises(subprocess.CalledProcessError) as e:
        subprocess.check_call([poolcounter_path, '-l', 'bad'])
    assert e.value.returncode == 1



import re


def test_uptime(poolcounter, clients):
    """UPTIME fetches uptime"""
    client = clients.get(1)
    client.send('STATS UPTIME')
    recv = client.receive()
    assert re.search(r'uptime: \d+ days, \d+h \d+m \d+s', recv)


def test_full(poolcounter, clients):
    """FULL fetches lots of stuff"""
    client = clients.get(1)
    client.send('STATS FULL')
    recv = client.receive()
    assert re.search(r'uptime: \d+ days, \d+h \d+m \d+s', recv)
    assert re.search(
        r'total processing time: (((\d+ days )?\d+h )?\d+m )?\d+(\.\d+)?s',
        recv)
    assert re.search(
        r'average processing time: (((\d+ days )?\d+h )?\d+m )?\d+(\.\d+)?s',
        recv)
    assert re.search(
        r'gained time: (((\d+ days )?\d+h )?\d+m )?\d+(\.\d+)?s', recv)
    assert re.search(
        r'waiting time: (((\d+ days )?\d+h )?\d+m )?\d+(\.\d+)?s', recv)
    assert re.search(
        r'waiting time for me: (((\d+ days )?\d+h )?\d+m )?\d+(\.\d+)?s', recv)
    assert re.search(
        r'waiting time for anyone: (((\d+ days )?\d+h )?\d+m )?\d+(\.\d+)?s',
        recv)
    assert re.search(
        r'waiting time for good: (((\d+ days )?\d+h )?\d+m )?\d+(\.\d+)?s',
        recv)
    assert re.search(
        r'wasted timeout time: (((\d+ days )?\d+h )?\d+m )?\d+(\.\d+)?s', recv)
    assert re.search(r'total_acquired: \d+', recv)
    assert re.search(r'total_releases: \d+', recv)
    assert re.search(r'hashtable_entries: \d+', recv)
    assert re.search(r'processing_workers: \d+', recv)
    assert re.search(r'waiting_workers: \d+', recv)
    assert re.search(r'connect_errors: \d+', recv)
    assert re.search(r'failed_sends: \d+', recv)
    assert re.search(r'full_queues: \d+', recv)
    assert re.search(r'lock_mismatch: \d+', recv)
    assert re.search(r'lock_while_waiting: \d+', recv)
    assert re.search(r'release_mismatch: \d+', recv)
    assert re.search(r'processed_count: \d+', recv)


def test_waiting_workers(poolcounter, clients):
    """Waiting_workers is 0 because we aren't running concurrent tests"""
    client = clients.get(1)
    client.send('STATS waiting_workers')
    assert client.receive() == 'waiting_workers: 0'


def test_hashtable_entries(poolcounter, clients):
    """Hashtable_entries is 0 because we aren't running concurrent tests"""
    client = clients.get(1)
    client.send('STATS hashtable_entries')
    assert client.receive() == 'hashtable_entries: 0'



import pytest
import socket


def test_both_lock(poolcounter, clients, lock_type):
    """Both lock if there are enough workers and enough queue space"""
    client1, client2 = clients.get(2)
    client1.send('%s l 2 2 1' % lock_type)
    assert client1.receive() == 'LOCKED'
    client2.send('%s l 2 2 1' % lock_type)
    assert client2.receive() == 'LOCKED'


def test_only_one_lock_no_queue_space(poolcounter, clients, lock_type):
    """Only one can lock if there are enough workers but not enough
    queue space"""
    client1, client2 = clients.get(2)
    client1.send('%s l 2 1 1' % lock_type)
    assert client1.receive() == 'LOCKED'
    client2.send('%s l 2 1 1' % lock_type)
    assert client2.receive() == 'QUEUE_FULL'


def test_only_one_lock_no_workers_nor_queue_space(
        poolcounter, clients, lock_type):
    """Only one can lock if there aren't enough workers nor enough queue
    space"""
    client1, client2 = clients.get(2)
    client1.send('%s l 1 1 1' % lock_type)
    assert client1.receive() == 'LOCKED'
    client2.send('%s l 1 1 1' % lock_type)
    assert client2.receive() == 'QUEUE_FULL'


def test_close_socket_release(poolcounter, clients):
    """Closing a socket releases the lock"""
    client1, client2 = clients.get(2)
    client1.send('ACQ4ANY l 2 1 1')
    assert client1.receive() == 'LOCKED'
    # Close client1
    client1.close()
    client2.send('ACQ4ANY l 2 1 1')
    assert client2.receive() == 'LOCKED'


def test_second_timeout_gt_zero(poolcounter, clients):
    """Second client TIMEOUTs if first never unlocks and timeout > 0"""
    client1, client2 = clients.get(2)
    client1.send('ACQ4ME l 1 2 1')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ME l 1 2 1')
    assert client2.receive() == 'TIMEOUT'


def test_second_timeout_eq_zero(poolcounter, clients):
    """Second client TIMEOUTs if first never unlocks and timeout == 0"""
    client1, client2 = clients.get(2)
    client1.send('ACQ4ME l 1 2 0')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ME l 1 2 0')
    assert client2.receive() == 'TIMEOUT'


def test_second_timeout_after_first_finish(poolcounter, clients):
    """Second client TIMEOUTs if first never unlocks but can get lock
    after first finishes"""
    client1, client2 = clients.get(2)
    client1.send('ACQ4ME l 1 2 1')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ME l 1 2 1')
    assert client2.receive() == 'TIMEOUT'
    client2.send('ACQ4ME l 1 2 1')
    assert client2.receive() == 'TIMEOUT'
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    client2.send('ACQ4ME l 1 2 1')
    assert client2.receive() == 'LOCKED'


def test_second_timeout_after_first_finish_zero_timeout(poolcounter, clients):
    """Second client TIMEOUTs if first never unlocks but can get lock
    after first finishes and timeout == 0"""
    client1, client2 = clients.get(2)
    client1.send('ACQ4ME l 1 2 0')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ME l 1 2 0')
    assert client2.receive() == 'TIMEOUT'
    client2.send('ACQ4ME l 1 2 0')
    assert client2.receive() == 'TIMEOUT'
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    client2.send('ACQ4ME l 1 2 0')
    assert client2.receive() == 'LOCKED'


def test_second_lock_first_unlock(poolcounter, clients):
    """Second client's ACQ4ME LOCKs if first unlocks"""
    client1, client2 = clients.get(2)
    client1.send('ACQ4ME l 1 2 5')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ME l 1 2 5')
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    assert client2.receive() == 'LOCKED'


def test_second_lock_first_unlock_any(poolcounter, clients):
    """Second client's ACQ4ANY DONEs if first unlocks"""
    client1, client2 = clients.get(2)
    client1.send('ACQ4ANY l 1 2 5')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ANY l 1 2 5')
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    assert client2.receive() == 'DONE'


def test_done_no_unlock(poolcounter, clients):
    """DONE doesn't require unlock"""
    client1, client2 = clients.get(2)
    client1.send('ACQ4ME l 1 2 5')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ANY l 1 2 5')
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    assert client2.receive() == 'DONE'
    client1.send('ACQ4ME l 1 2 5')
    assert client1.receive() == 'LOCKED'


def test_done_no_consume_lock(poolcounter, clients):
    """DONE doesn't consume a lock"""
    client1, client2 = clients.get(2)
    client1.send('ACQ4ME l 1 5 5')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ANY l 1 5 5')
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    assert client2.receive() == 'DONE'
    client1.send('ACQ4ME l 1 5 5')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ANY l 1 5 5')
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    assert client2.receive() == 'DONE'
    client1.send('ACQ4ME l 1 5 5')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ANY l 1 5 5')
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    assert client2.receive() == 'DONE'
    client1.send('ACQ4ME l 1 5 5')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ANY l 1 5 5')
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    assert client2.receive() == 'DONE'
    client2.send('ACQ4ME l 1 2 5')
    assert client2.receive() == 'LOCKED'


def test_lock_twice_multiple_executors(poolcounter, clients, lock_type):
    """locking the same lock twice takes multiple executors"""
    client1, client2 = clients.get(2)
    client1.send('%s l 2 2 0' % lock_type)
    assert client1.receive() == 'LOCKED'
    client1.send('%s l 2 2 0' % lock_type)
    assert client1.receive() == 'LOCKED'
    client2.send('%s l 2 2 0' % lock_type)
    assert client2.receive() == 'QUEUE_FULL'


def test_disconnect_while_timeout(poolcounter, clients, lock_type):
    """Disconnecting while waiting for a timeout is OK"""
    client1, client2 = clients.get(2)
    client1.send('%s l 1 10 1' % lock_type)
    assert client1.receive() == 'LOCKED'
    client2.send('%s l 1 10 10' % lock_type)
    client2.close()
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'


def test_no_timeout_after_initial_delay(poolcounter, clients):
    """Do not get timeout after locking after initial delay"""
    client1, client2 = clients.get(2)
    client1.send('ACQ4ME l 1 10 1')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ME l 1 10 3')
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    assert client2.receive() == 'LOCKED'
    # assert no response
    with pytest.raises(socket.timeout):
        client2.receive()


def test_no_early_timeout_after_initial_delay(poolcounter, clients):
    """
    Do not get early timeout after initial delay

    The "gets no response" here is within 5 seconds, meaning the
    timeout doesn't come early
    """
    client1, client2 = clients.get(2)
    client1.send('ACQ4ME l 1 10 1')
    assert client1.receive() == 'LOCKED'
    client2.send('ACQ4ME l 1 10 3')
    client1.send('RELEASE')
    assert client1.receive() == 'RELEASED'
    assert client2.receive() == 'LOCKED'
    client2.send('ACQ4ME l 1 10 10')
    # assert no response
    with pytest.raises(socket.timeout):
        client2.receive()
